/**
 * GMyth Library
 *
 * @file gmyth/gmyth_context.c
 * 
 * @brief <p> GMythContext class contains general attributes and functions
 * that express the connection state of each mythtvfrontend.
 *
 * Copyright (C) 2006 INdT - Instituto Nokia de Tecnologia.
 * @author Hallyson Luiz de Morais Melo <hallyson.melo@indt.org.br>
 *
 *//*
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gmyth_context.h"

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>

#include "gmyth_query.h"
#include "gmyth_socket.h"

static void gmyth_context_class_init          (GMythContextClass *klass);
static void gmyth_context_init                (GMythContext *object);

static void gmyth_context_dispose  (GObject *object);
static void gmyth_context_finalize (GObject *object);


G_DEFINE_TYPE(GMythContext, gmyth_context, G_TYPE_OBJECT)

static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

static GMythContext *gcontext = NULL; 

static void
gmyth_context_class_init (GMythContextClass *klass)
{
	GObjectClass *gobject_class;

	gobject_class = (GObjectClass *) klass;

	gobject_class->dispose  = gmyth_context_dispose;
	gobject_class->finalize = gmyth_context_finalize;	
}

static void
gmyth_context_init (GMythContext *gmyth_context)
{

}

static void
gmyth_context_dispose  (GObject *object)
{
	GMythContext *gmyth_context = GMYTH_CONTEXT(object);

	if ( gmyth_context->gmyth_settings != NULL ) {
		g_object_unref (gmyth_context->gmyth_settings);	
		gmyth_context->gmyth_settings = NULL;
	}

	if ( gmyth_context->localhostname != NULL ) {
		g_string_free ( gmyth_context->localhostname, TRUE );
		gmyth_context->localhostname = NULL;
	}

	if (gmyth_context->server_socket != NULL) {
		g_object_unref (gmyth_context->server_socket);
		gmyth_context->server_socket = NULL;
	}

	G_OBJECT_CLASS (gmyth_context_parent_class)->dispose (object);
}

static void
gmyth_context_finalize (GObject *object)
{
	g_signal_handlers_destroy (object);

	G_OBJECT_CLASS (gmyth_context_parent_class)->finalize (object);
}

/** Gets the some important address translation info, 
 * from the client socket that will open a connection.
 *
 * @return gint error number
 */
static gint
myth_context_toaddrinfo( gchar *addr, gint port, struct addrinfo **addrInfo )
{
	struct addrinfo hints;
	gchar *portStr = g_strnfill( 32, ' ' );
	gint errorn = EADDRNOTAVAIL;

	memset( &hints, 0, sizeof(hints) );
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	/* hints.ai_flags = AI_NUMERICHOST; */
	if ( port != -1 )
		sprintf(portStr, "%d", port);
	else
		portStr = NULL;

	g_debug( "[%s] Address: %s, port: %s\n", __FUNCTION__, addr, portStr );
	if ( ( errorn = getaddrinfo(addr, portStr, &hints, addrInfo) ) != 0 ) {
		g_printerr( "[%s] Socket ERROR: %s\n", __FUNCTION__, gai_strerror(errorn) );
	}

	return errorn;

}

/** Initializes the GMythContext object. It reads local system information
 * load gmyth user settings and start connection to the database.
 * 
 * @return TRUE if initialization was successfull, 
 * FALSE if any error happens.
 */
gboolean
gmyth_context_initialize ()
{
	GString *localhost = NULL;

	if (gcontext == NULL) {
		gcontext = GMYTH_CONTEXT ( g_object_new (GMYTH_CONTEXT_TYPE, FALSE));
	}

	localhost = gmyth_socket_get_local_hostname( );

	if (localhost==NULL || localhost->len <=0 ) {
		g_warning ("[%s] Could not determine local hostname. Setting to 127.0.0.1", __FUNCTION__);
		gcontext->localhostname = g_string_new ("127.0.0.1");
	} else {
		gcontext->localhostname = localhost;
	}

	if (gcontext->gmyth_settings) {
		g_object_unref (gcontext->gmyth_settings);	
		gcontext->gmyth_settings = NULL;
	}

	gcontext->gmyth_settings = gmyth_settings_new ();
	if (!gmyth_settings_load (gcontext->gmyth_settings)) {
		g_warning ("GMythContext: Settings file not opened!\n");	
	} else {
		g_debug ("GMythContext: Settings file loaded");
	}

	if (gcontext->server_socket != NULL) {
		g_object_unref (gcontext->server_socket);
		gcontext->server_socket = NULL;
	}

	GString *server_hostname = 
        gmyth_settings_get_backend_hostname(gcontext->gmyth_settings);
	int server_port = gmyth_settings_get_backend_port(gcontext->gmyth_settings);

	gcontext->server_socket = gmyth_socket_new ();
	if (!gmyth_socket_connect_to_backend (gcontext->server_socket,
				server_hostname->str, server_port, FALSE)) {
		g_warning ("[%s] Socket connection to backend error!", __FUNCTION__);
		g_object_unref (gcontext->server_socket);
		gcontext->server_socket = NULL;
		return FALSE;	
	}

	return TRUE;
}

/** Formats a Mythtv protocol command based on strlist and sends it to
 * the connected backend. The backend response is overwritten into strlist.
 * 
 * @param strlist the string list to be sent, 
 * and on which the answer will be written.
 * @return TRUE if command was sent and an answer was received, FALSE if any
 * error happens.
 */
gboolean
gmyth_context_send_receive_stringlist (GMythStringList *strlist)
{
	gint ok = -1;

	if (!gcontext || !(gcontext->server_socket)) {
		g_warning ("[%s] GMythContext not initialized", __FUNCTION__);
		return FALSE;
	}

	//g_static_mutex_lock( &mutex );

	ok = gmyth_socket_sendreceive_stringlist (gcontext->server_socket, strlist);

	//g_static_mutex_unlock( &mutex );

	if (!ok) {
		g_warning ("Connection to backend server lost");
	}

	return (ok ? TRUE : FALSE);
}

/** Gets the GMythSettings object associated to this context.
 * 
 * @return The GMythSettings object currently valid or NULL if the settings
 * were not opened.
 */
GMythSettings*
gmyth_context_get_settings ()
{
	if (!gcontext) {
		g_warning ("[%s] GMythContext not initialized\n", __FUNCTION__);
		return NULL;
	}

	return gcontext->gmyth_settings;
}

/** Gets the machine local hostname.
 * 
 * @param hostname a valid GString object to be overwritten with the local
 * hostname.
 * @return true if the hostname was read, false if any error happened.
 */
gboolean
gmyth_context_get_local_hostname (GString *hostname)
{
	if (!hostname) {
		g_warning ("[%s] Received null argument", __FUNCTION__);
		return FALSE;
	}

	g_string_assign (hostname, gcontext->localhostname->str);

	return TRUE;
}

/** Gets a setting information from the backend mysql database.
 * 
 * @param key The setting key to be retrieved.
 * @param host the hostname associated to the desired setting.
 * @param default_value the default setting value if could not query from
 * backend database.
 * @return The setting value loaded from database, or the given default value
 * if the query fails.
 */
GString*
gmyth_context_get_setting_onhost (GString *key, GString *host, GString *default_value)
{
	GString *query_str;

	// TODO: Reuse msql query in gmyth_context
	GMythQuery *gmyth_query = gmyth_query_new ();

	if (gmyth_query_connect (gmyth_query)) {
		MYSQL_RES *msql_res;
		MYSQL_ROW msql_row;

		query_str = g_string_new ("");
		g_string_printf (query_str, "SELECT data FROM settings WHERE value = \"%s\" "
				"AND hostname = \"%s\" ;", key->str, host->str);

		msql_res = gmyth_query_process_statement (gmyth_query, query_str->str);
		if (msql_res) {
			msql_row = mysql_fetch_row (msql_res);
			if (msql_row != NULL) {
				return g_string_new (msql_row[0]);
			}
		}

		g_object_unref (gmyth_query);
	} else {
		g_warning ("Database not open while trying to load setting: %s", key->str);
	}

	return default_value;
}

/** Verify if the context is currently connected to a backend.
 * 
 * @return true if connection was opened, false if not.
 */
gboolean
gmyth_context_check_connection ()
{
	// FIXME: Check this based on socket states
	if (!gcontext) {
		g_warning ("[%s] GMythContext not initialized", __FUNCTION__);
		return FALSE;
	}

	return (gcontext->server_socket != NULL);
}
